@@ -702,7 +702,18 @@ escl_precheck_decode (const proto_ctx *ctx)
 
 /* Fix Location: URL
  *
- * Can be used as http_query_onredir() callback
+ * It replaces host part of uri with the host part from `orig_uri` and
+ * used for two purposes:
+ *   1) To fix URL in the Location: header. We don't trust the device
+ *      and only allow it to specify path, not host (host is preserved
+ *      from the `orig_uri` -- the URL used to initiate scanning
+ *   2) As redirection callback, together with the quirk_localhost.
+ *      At this case, Host: is purposely invalid, which makes
+ *      redirection unreliable (most likely, redirection will
+ *      use "localhost" from the Host: header as the host part
+ *      of URL).
+ *
+ * Can be directly used as http_query_onredir() callback
  */
 static void
 escl_scan_fix_location (void *p, http_uri *uri, const http_uri *orig_uri)
@@ -781,6 +792,17 @@ escl_scan_query (const proto_ctx *ctx)
     query = escl_http_query(ctx, "ScanJobs", "POST",
         xml_wr_finish_compact(xml));
 
+    /* Kyocera ECOSYS M6526cdn drops TLS connection after sending
+     * response HTTP headers, but before the body transfer is completed.
+     *
+     * As for this request we are only interested in the response
+     * headers, we can ignore this kind of error
+     *
+     * See here for details:
+     *   https://github.com/alexpevzner/sane-airscan/issues/163
+     */
+    http_query_no_need_response_body(query);
+
     /* It's a dirty hack
      *
      * HP LaserJet MFP M630, HP Color LaserJet FlowMFP M578 and
@@ -793,7 +815,7 @@ escl_scan_query (const proto_ctx *ctx)
      * Note, this hack doesn't work with some other printers
      * see #92, #98 for details
      */
-    if (escl->quirk_localhost) {
+    if (escl->quirk_localhost && !http_uri_is_loopback(ctx->base_uri)) {
         http_query_set_request_header(query, "Host", "localhost");
         http_query_onredir(query, escl_scan_fix_location);
     }
@@ -834,7 +856,19 @@ escl_scan_decode (const proto_ctx *ctx)
         goto ERROR;
     }
 
-    escl_scan_fix_location(NULL, uri, http_query_uri(ctx->query));
+    /* Don't trust hostname in Location, replace it with hostname
+     * from the ctx->query (which represents scan request)
+     *
+     * The initial idea behind this approach is that scanner may
+     * not have a strong knowledge of its own host name so hostname
+     * supplied by device may be inaccurate.
+     *
+     * At least one device, HP Deskjet 3520 series, has demonstrated
+     * this behavior when scanning via IPP over USB. Instead of (expected)
+     * localhost host name, it returns CN26F178X605SZ.03f011b0.hpLedmUSB
+     */
+    http_uri_fix_host(uri, http_query_uri(ctx->query), NULL);
+
     result.data.location = str_dup(http_uri_str(uri));
     http_uri_free(uri);
 
