@@ -22,6 +22,7 @@
 #include <sys/un.h>
 #include <unistd.h>
 
+#include <avahi-common/domain.h>
 #include <gnutls/gnutls.h>
 
 /******************** Constants ********************/
@@ -863,7 +864,7 @@ http_uri_str (http_uri *uri)
 /* Get URI's host address. If Host address is not literal, returns NULL
  */
 const struct sockaddr*
-http_uri_addr (http_uri *uri)
+http_uri_addr (const http_uri *uri)
 {
     if (uri->addr.sockaddr.sa_family == AF_UNSPEC) {
         return NULL;
@@ -873,6 +874,9 @@ http_uri_addr (http_uri *uri)
 }
 
 /* Get URI path
+ *
+ * Note, if URL has empty path (i.e., "http://1.2.3.4"), the
+ * empty string will be returned
  */
 const char*
 http_uri_get_path (const http_uri *uri)
@@ -904,6 +908,76 @@ http_uri_get_host (const http_uri *uri)
     return uri->host;
 }
 
+/* http_uri_host_is checks if URI's host name is equal to the
+ * specified string.
+ *
+ * It does its best to compare domain names correctly, taking
+ * in account only significant difference (for example, the difference
+ * in upper/lower case * in domain names is not significant).
+ */
+bool
+http_uri_host_is (const http_uri *uri, const char *host)
+{
+    const char *uri_host = http_uri_get_host(uri);
+
+    if (http_uri_host_is_literal(uri)) {
+        return !strcasecmp(uri_host, host);
+    }
+
+    if (!avahi_is_valid_domain_name(uri_host)) {
+        return false;
+    }
+
+    if (!avahi_is_valid_domain_name(host)) {
+        return false;
+    }
+
+    return avahi_domain_equal(uri_host, host);
+}
+
+/* http_uri_host_is_literal returns true if URI uses literal
+ * IP address
+ */
+bool
+http_uri_host_is_literal (const http_uri *uri)
+{
+    return http_uri_addr(uri) != NULL;
+}
+
+/* Set URI host into the literal IP address.
+ */
+void
+http_uri_set_host_addr (http_uri *uri, ip_addr addr)
+{
+    ip_straddr straddr = ip_addr_to_straddr(addr, true);
+    char       *host = straddr.text;
+    char       *s;
+
+    /* Remove square brackets around IPv6 address */
+    if (host[0] == '[' && host[strlen(host) - 1] == ']') {
+        host[strlen(host) - 1] = '\0';
+        host ++;
+    }
+
+    /* Escape % character (zone suffix delimiter) */
+    s = strchr(host, '%');
+    if (s != NULL) {
+        /* Note, we replace '%' with '%25' plus don't forget
+         * the terminating '\0'. Hence strlen(host) + 3
+         */
+        char   *host2 = alloca(strlen(host) + 3);
+        size_t sz = s - host;
+
+        memcpy(host2, host, sz);
+        memcpy(host2 + sz, "%25", 3);
+        strcpy(host2 + sz + 3, s + 1);
+        host = host2;
+    }
+
+    /* Set the host */
+    http_uri_field_replace(uri, UF_HOST, host);
+}
+
 /* Fix URI host: if `match` is NULL or uri's host matches `match`,
  * replace uri's host and port with values taken from the base_uri
  */
@@ -1926,30 +2000,6 @@ http_client_timeout (http_client *client, int timeout)
     }
 }
 
-/* Cancel all pending queries with matching address family and uintptr
- */
-void
-http_client_cancel_af_uintptr (http_client *client, int af, uintptr_t uintptr)
-{
-    ll_head leftover;
-    ll_node *node;
-
-    ll_init(&leftover);
-
-    while ((node = ll_pop_beg(&client->pending)) != NULL) {
-        http_query *q = http_query_by_ll_node(node);
-
-        if (uintptr == http_query_get_uintptr(q) &&
-            af == http_uri_af(http_query_uri(q))) {
-            http_query_cancel(q);
-        } else {
-            ll_push_end(&leftover, node);
-        }
-    }
-
-    ll_cat(&client->pending, &leftover);
-}
-
 /* Check if client has pending queries
  */
 bool
@@ -1982,6 +2032,9 @@ struct http_query {
     eloop_timer       *timeout_timer;           /* Timeout timer */
     int               timeout_value;            /* In milliseconds */
 
+    /* Miscellaneous options */
+    bool              no_need_response_body;    /* Response body not needed */
+
     /* Low-level I/O */
     bool              submitted;                /* http_query_submit() called */
     uint64_t          eloop_callid;             /* For eloop_call_cancel */
@@ -2001,6 +2054,7 @@ struct http_query {
 
     /* HTTP parser */
     http_parser       http_parser;              /* HTTP parser structure */
+    bool              http_headers_received;    /* HTTP headers received */
     bool              http_parser_done;         /* Message parsing done */
 
     /* Data handling */
@@ -2065,6 +2119,7 @@ http_query_reset (http_query *q)
     str_trunc(q->rq_buf);
     q->rq_off = 0;
 
+    q->http_headers_received = false;
     q->http_parser_done = false;
 
     http_data_unref(q->response_data);
@@ -2131,7 +2186,7 @@ http_query_set_host (http_query *q)
             dport = -1;
         }
 
-        s = ip_straddr_from_sockaddr_dport(addr, dport, false);
+        s = ip_straddr_from_sockaddr_dport(addr, dport, false, true);
         http_query_set_request_header(q, "Host", s.text);
 
         return;
@@ -2265,6 +2320,20 @@ http_query_timeout (http_query *q, int timeout)
     }
 }
 
+/* Set 'no_need_response_body' flag
+ *
+ * This flag notifies, that http_query issued is only interested
+ * in the HTTP response headers, not body
+ *
+ * If this flag is set, after successful reception of response
+ * HTTP header, errors in fetching response body is ignored
+ */
+void
+http_query_no_need_response_body (http_query *q)
+{
+    q->no_need_response_body = true;
+}
+
 /* Cancel query timeout timer
  */
 static void
@@ -2405,6 +2474,29 @@ http_query_redirect (http_query *q, const char *method)
     return NULL;
 }
 
+/* Check if I/O error should be ignored for this query
+ */
+static bool
+http_query_ignore_error (http_query *q)
+{
+    int status;
+
+    /* Headers must be received */
+    if (!q->http_headers_received) {
+        return false;
+    }
+
+    /* Depending on HTTP status, response body may be irrelevant */
+    status = q->http_parser.status_code;
+    switch (status / 100) {
+        case 1: case 3: case 4: case 5:
+            return true;
+    }
+
+    /* If `no_need_response_body' is set, ignore the error */
+    return q->no_need_response_body;
+}
+
 /* Complete query processing
  */
 static void
@@ -2418,6 +2510,13 @@ http_query_complete (http_query *q, error err)
     /* Unlink query from a client */
     ll_del(&q->chain);
 
+    /* In some cases, I/O error can be ignored. Check for that */
+    if (err != NULL && http_query_ignore_error(q)) {
+        log_debug(client->log, "HTTP %s %s: %s (ignored)", q->method,
+                http_uri_str(q->uri), ESTRING(err));
+        err = NULL;
+    }
+
     /* Issue log messages */
     q->err = err;
     if (err != NULL) {
@@ -2503,6 +2602,8 @@ http_query_on_headers_complete (http_parser *parser)
                 http_uri_str(q->uri),
                 http_query_status(q));
 
+        q->http_headers_received = true;
+
         if (q->onrxhdr != NULL) {
             q->onrxhdr(q->client->ptr, q);
         }
@@ -2597,6 +2698,8 @@ http_query_fdpoll_callback (int fd, void *data, ELOOP_FDPOLL_MASK mask)
 
         if (rc > 0) {
             log_debug(q->client->log, "HTTP %d bytes sent", (int) rc);
+            trace_hexdump(log_ctx_trace(q->client->log), '>',
+                q->rq_buf + q->rq_off, rc);
         }
 
         if (rc < 0) {
@@ -2640,6 +2743,7 @@ http_query_fdpoll_callback (int fd, void *data, ELOOP_FDPOLL_MASK mask)
         rc = http_query_sock_recv(q, io_buf, sizeof(io_buf));
         if (rc > 0) {
             log_debug(q->client->log, "HTTP %d bytes received", (int) rc);
+            trace_hexdump(log_ctx_trace(q->client->log), '<', io_buf, rc);
         }
 
         if (rc < 0) {
@@ -2877,6 +2981,7 @@ http_query_start_processing (void *p)
     http_query      *q = (http_query*) p;
     http_uri_field  field;
     char            *host, *port;
+    const char      *path;
     struct addrinfo hints;
     int             rc;
 
@@ -2945,10 +3050,18 @@ http_query_start_processing (void *p)
         http_query_set_host(q);
     }
 
+    /* Obtain path. Note, URL format allows path to be empty,
+     * while HTTP request requires non-empty string
+     */
+    path = http_uri_get_path(q->uri);
+    if (*path == '\0') {
+        path = "/";
+    }
+
     /* Format HTTP request */
     str_trunc(q->rq_buf);
     q->rq_buf = str_append_printf(q->rq_buf, "%s %s HTTP/1.1\r\n",
-        q->method, http_uri_get_path(q->uri));
+        q->method, path);
 
     if (q->request_data != NULL) {
         char buf[64];
