@@ -8,6 +8,10 @@
 
 #include "airscan.h"
 
+/* Saved init flags
+ */
+static AIRSCAN_INIT_FLAGS airscan_init_flags;
+
 /* Initialize airscan
  */
 SANE_Status
@@ -15,6 +19,9 @@ airscan_init (AIRSCAN_INIT_FLAGS flags, const char *log_msg)
 {
     SANE_Status status;
 
+    /* Save init flags */
+    airscan_init_flags = flags;
+
     /* Initialize logging -- do it early */
     log_init();
     trace_init();
@@ -83,5 +90,13 @@ airscan_cleanup (const char *log_msg)
     log_cleanup(); /* Must be the last thing to do */
 }
 
+/* Get init flags from the airscan_init call
+ */
+AIRSCAN_INIT_FLAGS
+airscan_get_init_flags (void)
+{
+    return airscan_init_flags;
+}
+
 /* vim:ts=8:sw=4:et
  */
