@@ -16,9 +16,9 @@
 #include <string.h>
 
 /******************** Constants *********************/
-/* Max time to wait until device table is ready, in seconds
+/* Max time to wait until device table is ready, in milliseconds
  */
-#define ZEROCONF_READY_TIMEOUT                  5
+#define ZEROCONF_READY_TIMEOUT                  5000
 
 /******************** Local Types *********************/
 /* zeroconf_device represents a single device
@@ -447,7 +447,7 @@ zeroconf_merge_check (zeroconf_device *device, zeroconf_finding *finding)
     return false;
 }
 
-/* Find device, suitable for merging with specified findind
+/* Find device, suitable for merging with specified finding
  */
 static zeroconf_device*
 zeroconf_merge_find (zeroconf_finding *finding)
@@ -576,7 +576,7 @@ zeroconf_endpoint_copy_single (const zeroconf_endpoint *endpoint)
 
 /* Free single zeroconf_endpoint
  */
-static void
+void
 zeroconf_endpoint_free_single (zeroconf_endpoint *endpoint)
 {
     http_uri_free(endpoint->uri);
@@ -746,6 +746,25 @@ zeroconf_endpoint_list_sort_dedup (zeroconf_endpoint *list)
     return list;
 }
 
+/* Check if list of endpoints already contains the given
+ * endpoint (i.e., endpoint with the same URI and protocol)
+ */
+bool
+zeroconf_endpoint_list_contains (const zeroconf_endpoint *list,
+        const zeroconf_endpoint *endpoint)
+{
+    while (list != NULL) {
+        if (list->proto == endpoint->proto &&
+            http_uri_equal(list->uri, endpoint->uri)) {
+            return true;
+        }
+
+        list = list->next;
+    }
+
+    return false;
+}
+
 /* Check if endpoints list contains a non-link-local address
  * of the specified address family
  */
@@ -808,6 +827,28 @@ zeroconf_find_static_by_ident (const char *ident)
     return NULL;
 }
 
+/***** Miscellaneous functions for zeroconf_finding *****/
+
+/* Compare two pointers to pointers to zeroconf_finding (zeroconf_finding**)
+ * by index+name, for qsort
+ */
+int
+zeroconf_finding_qsort_by_index_name (const void *p1, const void *p2)
+{
+    const zeroconf_finding *f1 = *(const zeroconf_finding * const *) p1;
+    const zeroconf_finding *f2 = *(const zeroconf_finding * const *) p2;
+
+    if (f1->ifindex < f2->ifindex) {
+        return -1;
+    }
+
+    if (f1->ifindex > f2->ifindex) {
+        return 1;
+    }
+
+    return strcmp(f1->name, f2->name);
+}
+
 /******************** Events from discovery providers *********************/
 /* Publish the zeroconf_finding.
  */
@@ -1335,7 +1376,7 @@ static void
 zeroconf_start_stop_callback (bool start)
 {
     if (start) {
-        zeroconf_initscan_timer = eloop_timer_new(ZEROCONF_READY_TIMEOUT * 1000,
+        zeroconf_initscan_timer = eloop_timer_new(ZEROCONF_READY_TIMEOUT,
                 zeroconf_initscan_timer_callback, NULL);
     } else {
         if (zeroconf_initscan_timer != NULL) {
